@@ -5,14 +5,16 @@
  *
  */
 
-#include <stdlib.h>
+#define _CRT_SECURE_NO_WARNINGS
+
+#include <signal.h>
 #include <stdbool.h>
 #include <stdio.h>
+#include <stdlib.h>
 #include <string.h>
-#include <signal.h>
 
-#include "cs101_slave.h"
 #include "cs101_file_service.h"
+#include "cs101_slave.h"
 
 #include "hal_serial.h"
 #include "hal_thread.h"
@@ -31,14 +33,13 @@ static uint8_t sectionData1[500];
 static uint8_t sectionData2[500];
 
 static uint64_t
-getFileDate (CS101_IFileProvider self)
+getFileDate(CS101_IFileProvider self)
 {
-
     return 0;
 }
 
 static int
-getFileSize (CS101_IFileProvider self)
+getFileSize(CS101_IFileProvider self)
 {
     printf("getFileSize --> %i\n", fileSize);
 
@@ -46,7 +47,7 @@ getFileSize (CS101_IFileProvider self)
 }
 
 static int
-getSectionSize (CS101_IFileProvider self, int sectionNumber)
+getSectionSize(CS101_IFileProvider self, int sectionNumber)
 {
     printf("getSectionSize(%i)\n", sectionNumber);
 
@@ -59,11 +60,11 @@ getSectionSize (CS101_IFileProvider self, int sectionNumber)
 }
 
 static bool
-getSegmentData (CS101_IFileProvider self, int sectionNumber, int offset, int size, uint8_t* data)
+getSegmentData(CS101_IFileProvider self, int sectionNumber, int offset, int size, uint8_t* data)
 {
     printf("getSegmentData(NoS=%i, offset=%i, size=%i)\n", sectionNumber, offset, size);
-    if (sectionNumber == 0) {
-
+    if (sectionNumber == 0)
+    {
         int i;
 
         for (i = 0; i < size; i++)
@@ -71,7 +72,8 @@ getSegmentData (CS101_IFileProvider self, int sectionNumber, int offset, int siz
 
         return true;
     }
-    else if (sectionNumber == 1) {
+    else if (sectionNumber == 1)
+    {
         int i;
 
         for (i = 0; i < size; i++)
@@ -84,7 +86,7 @@ getSegmentData (CS101_IFileProvider self, int sectionNumber, int offset, int siz
 }
 
 static void
-transferComplete (CS101_IFileProvider self, bool success)
+transferComplete(CS101_IFileProvider self, bool success)
 {
     printf("FILE TRANSFER COMPLETE\n");
 }
@@ -106,26 +108,29 @@ initializeFiles()
     fileProvider[0].transferComplete = transferComplete;
 
     int i;
-    for (i = 0; i < fileSize / 2; i++) {
-        sectionData1[i] = (uint8_t) (i % 0x100);
-        sectionData2[i] = (uint8_t) ((i + 1) % 0x100);
+    for (i = 0; i < fileSize / 2; i++)
+    {
+        sectionData1[i] = (uint8_t)(i % 0x100);
+        sectionData2[i] = (uint8_t)((i + 1) % 0x100);
     }
 }
 
 static CS101_IFileProvider
-getNextFile (void* parameter, CS101_IFileProvider continueAfter)
+getNextFile(void* parameter, CS101_IFileProvider continueAfter)
 {
     return NULL;
 }
 
 static CS101_IFileProvider
-getFile (void* parameter, int ca, int ioa, uint16_t nof, int* errCode)
+getFile(void* parameter, int ca, int ioa, uint16_t nof, int* errCode)
 {
     printf("getFile %i:%i (type:%i)\n", ca, ioa, nof);
 
     int i;
-    for (i = 0; i < numberOfFiles; i++) {
-        if ((ca == fileProvider[i].ca) && (ioa == fileProvider[i].ioa)) {
+    for (i = 0; i < numberOfFiles; i++)
+    {
+        if ((ca == fileProvider[i].ca) && (ioa == fileProvider[i].ioa))
+        {
             return &(fileProvider[i]);
         }
     }
@@ -136,11 +141,12 @@ getFile (void* parameter, int ca, int ioa, uint16_t nof, int* errCode)
 static void
 IFileReceiver_finished(CS101_IFileReceiver self, CS101_FileErrorCode result)
 {
-    FILE* fp = (FILE*) self->object;
+    FILE* fp = (FILE*)self->object;
 
     fclose(fp);
 
-    if (result != CS101_FILE_ERROR_SUCCESS) {
+    if (result != CS101_FILE_ERROR_SUCCESS)
+    {
         remove("upload.dat");
     }
 
@@ -148,9 +154,9 @@ IFileReceiver_finished(CS101_IFileReceiver self, CS101_FileErrorCode result)
 }
 
 static void
-IFileReceiver_segmentReceived (CS101_IFileReceiver self, uint8_t sectionName, int offset, int size, uint8_t* data)
+IFileReceiver_segmentReceived(CS101_IFileReceiver self, uint8_t sectionName, int offset, int size, uint8_t* data)
 {
-    FILE* fp = (FILE*) self->object;
+    FILE* fp = (FILE*)self->object;
 
     printf("File upload - section %i - offset: %i - size: %i\n", sectionName, offset, size);
 
@@ -159,25 +165,35 @@ IFileReceiver_segmentReceived (CS101_IFileReceiver self, uint8_t sectionName, in
 
 struct sCS101_IFileReceiver myFileReceiver;
 
-
-
 static CS101_IFileReceiver
-fileReadyHandler (void* parameter, int ca, int ioa, uint16_t nof, int lengthOfFile, int* err)
+fileReadyHandler(void* parameter, int ca, int ioa, uint16_t nof, int lengthOfFile, int* err)
 {
-    if ((ca == 1) && (ioa == 30001)) {
-
+    if ((ca == 1) && (ioa == 30001))
+    {
         myFileReceiver.object = fopen("upload.dat", "wb");
 
-        myFileReceiver.finished = IFileReceiver_finished;
-        myFileReceiver.segmentReceived = IFileReceiver_segmentReceived;
+        if (myFileReceiver.object == NULL)
+        {
+            printf("Rejected file upload - cannot create local file\n");
 
-        printf("Accepted file upload\n");
+            *err = 1;
 
-        *err = 0;
+            return NULL;
+        }
+        else
+        {
+            myFileReceiver.finished = IFileReceiver_finished;
+            myFileReceiver.segmentReceived = IFileReceiver_segmentReceived;
 
-        return &myFileReceiver;
+            printf("Accepted file upload\n");
+
+            *err = 0;
+
+            return &myFileReceiver;
+        }
     }
-    else {
+    else
+    {
         printf("Rejected file upload - unknown file\n");
 
         *err = 1;
@@ -186,21 +202,17 @@ fileReadyHandler (void* parameter, int ca, int ioa, uint16_t nof, int lengthOfFi
     }
 }
 
-
 void
 printCP56Time2a(CP56Time2a time)
 {
-    printf("%02i:%02i:%02i %02i/%02i/%04i", CP56Time2a_getHour(time),
-                             CP56Time2a_getMinute(time),
-                             CP56Time2a_getSecond(time),
-                             CP56Time2a_getDayOfMonth(time),
-                             CP56Time2a_getMonth(time),
-                             CP56Time2a_getYear(time) + 2000);
+    printf("%02i:%02i:%02i %02i/%02i/%04i", CP56Time2a_getHour(time), CP56Time2a_getMinute(time),
+           CP56Time2a_getSecond(time), CP56Time2a_getDayOfMonth(time), CP56Time2a_getMonth(time),
+           CP56Time2a_getYear(time) + 2000);
 }
 
 /* Callback handler to log sent or received messages (optional) */
 static void
-rawMessageHandler (void* parameter, uint8_t* msg, int msgSize, bool sent)
+rawMessageHandler(void* parameter, uint8_t* msg, int msgSize, bool sent)
 {
     if (sent)
         printf("SEND: ");
@@ -208,7 +220,8 @@ rawMessageHandler (void* parameter, uint8_t* msg, int msgSize, bool sent)
         printf("RCVD: ");
 
     int i;
-    for (i = 0; i < msgSize; i++) {
+    for (i = 0; i < msgSize; i++)
+    {
         printf("%02x ", msg[i]);
     }
 
@@ -217,9 +230,11 @@ rawMessageHandler (void* parameter, uint8_t* msg, int msgSize, bool sent)
 
 /* Callback handler that is called when a clock synchronization command is received */
 static bool
-clockSyncHandler (void* parameter, IMasterConnection connection, CS101_ASDU asdu, CP56Time2a newTime)
+clockSyncHandler(void* parameter, IMasterConnection connection, CS101_ASDU asdu, CP56Time2a newTime)
 {
-    printf("Process time sync command with time "); printCP56Time2a(newTime); printf("\n");
+    printf("Process time sync command with time ");
+    printCP56Time2a(newTime);
+    printf("\n");
 
     return true;
 }
@@ -230,7 +245,8 @@ interrogationHandler(void* parameter, IMasterConnection connection, CS101_ASDU a
 {
     printf("Received interrogation for group %i\n", qoi);
 
-    if (qoi == 20) { /* only handle station interrogation */
+    if (qoi == 20)
+    { /* only handle station interrogation */
 
         CS101_AppLayerParameters alParams = IMasterConnection_getApplicationLayerParameters(connection);
 
@@ -238,18 +254,17 @@ interrogationHandler(void* parameter, IMasterConnection connection, CS101_ASDU a
 
         /* The CS101 specification only allows information objects without timestamp in GI responses */
 
-        CS101_ASDU newAsdu = CS101_ASDU_create(alParams, false, CS101_COT_INTERROGATED_BY_STATION,
-                0, 1, false, false);
+        CS101_ASDU newAsdu = CS101_ASDU_create(alParams, false, CS101_COT_INTERROGATED_BY_STATION, 0, 1, false, false);
 
-        InformationObject io = (InformationObject) MeasuredValueScaled_create(NULL, 100, -1, IEC60870_QUALITY_GOOD);
+        InformationObject io = (InformationObject)MeasuredValueScaled_create(NULL, 100, -1, IEC60870_QUALITY_GOOD);
 
         CS101_ASDU_addInformationObject(newAsdu, io);
 
-        CS101_ASDU_addInformationObject(newAsdu, (InformationObject)
-            MeasuredValueScaled_create((MeasuredValueScaled) io, 101, 23, IEC60870_QUALITY_GOOD));
+        CS101_ASDU_addInformationObject(newAsdu, (InformationObject)MeasuredValueScaled_create(
+                                                     (MeasuredValueScaled)io, 101, 23, IEC60870_QUALITY_GOOD));
 
-        CS101_ASDU_addInformationObject(newAsdu, (InformationObject)
-            MeasuredValueScaled_create((MeasuredValueScaled) io, 102, 2300, IEC60870_QUALITY_GOOD));
+        CS101_ASDU_addInformationObject(newAsdu, (InformationObject)MeasuredValueScaled_create(
+                                                     (MeasuredValueScaled)io, 102, 2300, IEC60870_QUALITY_GOOD));
 
         InformationObject_destroy(io);
 
@@ -257,15 +272,14 @@ interrogationHandler(void* parameter, IMasterConnection connection, CS101_ASDU a
 
         CS101_ASDU_destroy(newAsdu);
 
-        newAsdu = CS101_ASDU_create(alParams, false, CS101_COT_INTERROGATED_BY_STATION,
-                    0, 1, false, false);
+        newAsdu = CS101_ASDU_create(alParams, false, CS101_COT_INTERROGATED_BY_STATION, 0, 1, false, false);
 
-        io = (InformationObject) SinglePointInformation_create(NULL, 104, true, IEC60870_QUALITY_GOOD);
+        io = (InformationObject)SinglePointInformation_create(NULL, 104, true, IEC60870_QUALITY_GOOD);
 
         CS101_ASDU_addInformationObject(newAsdu, io);
 
-        CS101_ASDU_addInformationObject(newAsdu, (InformationObject)
-            SinglePointInformation_create((SinglePointInformation) io, 105, false, IEC60870_QUALITY_GOOD));
+        CS101_ASDU_addInformationObject(newAsdu, (InformationObject)SinglePointInformation_create(
+                                                     (SinglePointInformation)io, 105, false, IEC60870_QUALITY_GOOD));
 
         InformationObject_destroy(io);
 
@@ -273,17 +287,24 @@ interrogationHandler(void* parameter, IMasterConnection connection, CS101_ASDU a
 
         CS101_ASDU_destroy(newAsdu);
 
-        newAsdu = CS101_ASDU_create(alParams, true, CS101_COT_INTERROGATED_BY_STATION,
-                0, 1, false, false);
+        newAsdu = CS101_ASDU_create(alParams, true, CS101_COT_INTERROGATED_BY_STATION, 0, 1, false, false);
 
-        CS101_ASDU_addInformationObject(newAsdu, io = (InformationObject) SinglePointInformation_create(NULL, 300, true, IEC60870_QUALITY_GOOD));
-        CS101_ASDU_addInformationObject(newAsdu, (InformationObject) SinglePointInformation_create((SinglePointInformation) io, 301, false, IEC60870_QUALITY_GOOD));
-        CS101_ASDU_addInformationObject(newAsdu, (InformationObject) SinglePointInformation_create((SinglePointInformation) io, 302, true, IEC60870_QUALITY_GOOD));
-        CS101_ASDU_addInformationObject(newAsdu, (InformationObject) SinglePointInformation_create((SinglePointInformation) io, 303, false, IEC60870_QUALITY_GOOD));
-        CS101_ASDU_addInformationObject(newAsdu, (InformationObject) SinglePointInformation_create((SinglePointInformation) io, 304, true, IEC60870_QUALITY_GOOD));
-        CS101_ASDU_addInformationObject(newAsdu, (InformationObject) SinglePointInformation_create((SinglePointInformation) io, 305, false, IEC60870_QUALITY_GOOD));
-        CS101_ASDU_addInformationObject(newAsdu, (InformationObject) SinglePointInformation_create((SinglePointInformation) io, 306, true, IEC60870_QUALITY_GOOD));
-        CS101_ASDU_addInformationObject(newAsdu, (InformationObject) SinglePointInformation_create((SinglePointInformation) io, 307, false, IEC60870_QUALITY_GOOD));
+        CS101_ASDU_addInformationObject(
+            newAsdu, io = (InformationObject)SinglePointInformation_create(NULL, 300, true, IEC60870_QUALITY_GOOD));
+        CS101_ASDU_addInformationObject(newAsdu, (InformationObject)SinglePointInformation_create(
+                                                     (SinglePointInformation)io, 301, false, IEC60870_QUALITY_GOOD));
+        CS101_ASDU_addInformationObject(newAsdu, (InformationObject)SinglePointInformation_create(
+                                                     (SinglePointInformation)io, 302, true, IEC60870_QUALITY_GOOD));
+        CS101_ASDU_addInformationObject(newAsdu, (InformationObject)SinglePointInformation_create(
+                                                     (SinglePointInformation)io, 303, false, IEC60870_QUALITY_GOOD));
+        CS101_ASDU_addInformationObject(newAsdu, (InformationObject)SinglePointInformation_create(
+                                                     (SinglePointInformation)io, 304, true, IEC60870_QUALITY_GOOD));
+        CS101_ASDU_addInformationObject(newAsdu, (InformationObject)SinglePointInformation_create(
+                                                     (SinglePointInformation)io, 305, false, IEC60870_QUALITY_GOOD));
+        CS101_ASDU_addInformationObject(newAsdu, (InformationObject)SinglePointInformation_create(
+                                                     (SinglePointInformation)io, 306, true, IEC60870_QUALITY_GOOD));
+        CS101_ASDU_addInformationObject(newAsdu, (InformationObject)SinglePointInformation_create(
+                                                     (SinglePointInformation)io, 307, false, IEC60870_QUALITY_GOOD));
 
         InformationObject_destroy(io);
 
@@ -293,7 +314,8 @@ interrogationHandler(void* parameter, IMasterConnection connection, CS101_ASDU a
 
         IMasterConnection_sendACT_TERM(connection, asdu);
     }
-    else {
+    else
+    {
         IMasterConnection_sendACT_CON(connection, asdu, true);
     }
 
@@ -307,7 +329,7 @@ asduHandler(void* parameter, IMasterConnection connection, CS101_ASDU asdu)
     {
         printf("received single command\n");
 
-        if  (CS101_ASDU_getCOT(asdu) == CS101_COT_ACTIVATION)
+        if (CS101_ASDU_getCOT(asdu) == CS101_COT_ACTIVATION)
         {
             InformationObject io = CS101_ASDU_getElement(asdu, 0);
 
@@ -315,10 +337,10 @@ asduHandler(void* parameter, IMasterConnection connection, CS101_ASDU asdu)
             {
                 if (InformationObject_getObjectAddress(io) == 5000)
                 {
-                    SingleCommand sc = (SingleCommand) io;
+                    SingleCommand sc = (SingleCommand)io;
 
                     printf("IOA: %i switch to %i\n", InformationObject_getObjectAddress(io),
-                            SingleCommand_getState(sc));
+                           SingleCommand_getState(sc));
 
                     CS101_ASDU_setCOT(asdu, CS101_COT_ACTIVATION_CON);
                 }
@@ -344,7 +366,7 @@ resetCUHandler(void* parameter)
 {
     printf("Received reset CU\n");
 
-    CS101_Slave_flushQueues((CS101_Slave) parameter);
+    CS101_Slave_flushQueues((CS101_Slave)parameter);
 }
 
 static void
@@ -352,7 +374,8 @@ linkLayerStateChanged(void* parameter, int address, LinkLayerState state)
 {
     printf("Link layer state: ");
 
-    switch (state) {
+    switch (state)
+    {
     case LL_STATE_IDLE:
         printf("IDLE\n");
         break;
@@ -381,6 +404,12 @@ main(int argc, char** argv)
 
     SerialPort port = SerialPort_create(serialPort, 9600, 8, 'E', 1);
 
+    if (!port)
+    {
+        printf("Cannot open serial port %s\n", serialPort);
+        return 1;
+    }
+
     /* create a new slave/server instance with default link layer and application layer parameters */
     // CS101_Slave slave = CS101_Slave_create(port, NULL, NULL, IEC60870_LINK_LAYER_UNBALANCED);
     CS101_Slave slave = CS101_Slave_create(port, NULL, NULL, IEC60870_LINK_LAYER_BALANCED);
@@ -394,7 +423,6 @@ main(int argc, char** argv)
     LinkLayerParameters llParameters = CS101_Slave_getLinkLayerParameters(slave);
     llParameters->timeoutForAck = 500;
 
-
     /* set the callback handler for the clock synchronization command */
     CS101_Slave_setClockSyncHandler(slave, clockSyncHandler, NULL);
 
@@ -405,7 +433,7 @@ main(int argc, char** argv)
     CS101_Slave_setASDUHandler(slave, asduHandler, NULL);
 
     /* set handler for reset CU (reset communication unit) message */
-    CS101_Slave_setResetCUHandler(slave, resetCUHandler, (void*) slave);
+    CS101_Slave_setResetCUHandler(slave, resetCUHandler, (void*)slave);
 
     /* set timeout for detecting connection loss */
     CS101_Slave_setIdleTimeout(slave, 1500);
@@ -428,7 +456,7 @@ main(int argc, char** argv)
     CS101_Slave_addPlugin(slave, CS101_FileServer_getSlavePlugin(fileServer));
 
     /* uncomment to log messages */
-    //CS101_Slave_setRawMessageHandler(slave, rawMessageHandler, NULL);
+    // CS101_Slave_setRawMessageHandler(slave, rawMessageHandler, NULL);
 
     int16_t scaledValue = 0;
 
@@ -436,17 +464,18 @@ main(int argc, char** argv)
 
     SerialPort_open(port);
 
-    while (running) {
-
+    while (running)
+    {
         /* has to be called periodically */
         CS101_Slave_run(slave);
 
         /* Enqueue a measurement every second */
-        if (Hal_getTimeInMs() > (lastMessageSent + 1000)) {
-
+        if (Hal_getTimeInMs() > (lastMessageSent + 1000))
+        {
             CS101_ASDU newAsdu = CS101_ASDU_create(alParameters, false, CS101_COT_PERIODIC, 0, 1, false, false);
 
-            InformationObject io = (InformationObject) MeasuredValueScaled_create(NULL, 110, scaledValue, IEC60870_QUALITY_GOOD);
+            InformationObject io =
+                (InformationObject)MeasuredValueScaled_create(NULL, 110, scaledValue, IEC60870_QUALITY_GOOD);
 
             scaledValue++;
 
@@ -464,10 +493,10 @@ main(int argc, char** argv)
         Thread_sleep(1);
     }
 
-
-exit_program:
     CS101_Slave_destroy(slave);
 
     SerialPort_close(port);
     SerialPort_destroy(port);
+
+    return 0;
 }
